from ctypes import c_int
from ctypes import c_uint
from ctypes import c_uint32
from ctypes import c_int32
from ctypes import c_void_p
from ctypes import c_char_p
from ctypes import c_long
from ctypes import c_ulong
from ctypes import c_short
from ctypes import byref
from ctypes import memset
from ctypes import sizeof
from ctypes import Structure
from ctypes import POINTER
from ctypes import pointer
from ctypes import CDLL
from ctypes import get_errno

import os


""" Wrapper for ADC C library using ctypes
All ctypes functions have exactly the same names as the ones in the
C library. To each function corresponds python function that hides
C specific operations from the user. The names of these functions are
the sames as the ctypes functions without ADC prefix."""

class PyAdcConf(Structure):
    """
    Used to set and get configuration information

    :ivar type: The configuration type
    :ivar dev_type:
    :ivar route_to: route the configuration to the nth instance of "type"
    :ivar flags:
    :ivar mask: set to one the bit corresponding to the value you want
                to set or get
    :ivar value: Array of values, each corresponding to a specific
                 configuration option
    """
    __ADC_CONF_LEN = 64          # fixme what to do about it
    ADC_CONF_TYPE_BRD = 0
    ADC_CONF_TYPE_CHN = 1
    ADC_CONF_TYPE_ACQ = 2
    ADC_CONF_TYPE_CUS = 3
    ADC_CONF_TYPE_TRG_EXT = 4
    ADC_CONF_TYPE_TRG_THR = 5
    ADC_CONF_TYPE_TRG_TIM = 6
    __ADC_CONF_TYPE_LAST_INDEX = 7

    # BOARD configuration
    ADC_CONF_BRD_STATUS = 0
    ADC_CONF_BRD_MAX_FREQ_HZ = 1
    ADC_CONF_BRD_MIN_FREQ_HZ = 2
    ADC_CONF_BRD_STATE_MACHINE_STATUS = 3
    ADC_CONF_BRD_N_CHAN = 4
    ADC_CONF_BRD_N_TRG_EXT = 5
    ADC_CONF_BRD_N_TRG_THR = 6
    ADC_CONF_BRD_N_TRG_TIM = 7
    ADC_CONF_BRD_TEMPERATURE = 8
    ADC_CONF_UTC_TIMING_BASE_S = 9
    ADC_CONF_UTC_TIMING_BASE_T = 10
    ADC_CONF_UTC_TIMING_BASE_B = 11
    __ADC_CONF_BRD_ATTRIBUTE_LAST_INDEX = 12

    # CHANNEL configuration
    ADC_CONF_CHN_RANGE = 0
    ADC_CONF_CHN_TERMINATION = 1
    ADC_CONF_CHN_OFFSET = 2
    ADC_CONF_CHN_SATURATION = 3
    ADC_CONF_CHN_GAIN = 4
    __ADC_CONF_CHN_ATTRIBUTE_LAST_INDEX = 5

    # TRIGGER EXTERNAL configuration
    ADC_CONF_TRG_EXT_ENABLE = 0
    ADC_CONF_TRG_EXT_POLARITY = 1
    ADC_CONF_TRG_EXT_DELAY = 2
    __ADC_CONF_TRG_EXT_LAST_INDEX = 3

    # TRIGGER THRESHOLD configuration
    ADC_CONF_TRG_THR_ENABLE = 0
    ADC_CONF_TRG_THR_POLARITY = 1
    ADC_CONF_TRG_THR_DELAY = 2
    ADC_CONF_TRG_THR_THRESHOLD = 3
    ADC_CONF_TRG_THR_HYSTERESIS = 4
    __ADC_CONF_TRG_THR_LAST_INDEX = 5

    # TRIGGER TIMER configuration
    ADC_CONF_TRG_TIM_ENABLE = 0
    ADC_CONF_TRG_TIM_SECONDS = 1
    ADC_CONF_TRG_TIM_NANO_SECONDS = 2
    __ADC_CONF_TRG_TIM_LAST_INDEX = 1

    # ACQUISITION configuration
    ADC_CONF_ACQ_N_SHOTS = 0
    ADC_CONF_ACQ_POST_SAMP = 1
    ADC_CONF_ACQ_PRE_SAMP = 2
    ADC_CONF_ACQ_UNDERSAMPLE = 3
    ADC_CONF_ACQ_FREQ_HZ = 4
    ADC_CONF_ACQ_N_BITS = 5
    __ADC_CONF_ACQ_ATTRIBUTE_LAST_INDEX = 6

    _fields_ = [("type", c_int),
                ("dev_type", c_uint),
                ("route_to", c_uint),
                ("flags", c_uint),
                ("mask", c_ulong),
                ("value", c_uint32 * __ADC_CONF_LEN)]

    def __init__(self, config_type, route_to):
        """
        Initialize a configuration structure

        :ivar config_type: configuration type
        :ivar route_to: index
        """
        memset(byref(self), 0, sizeof(PyAdcConf))
        self.type = config_type
        self.route_to = route_to

    def mask_set(self, index):
        """
        Set the bit in the configuration mask

        :ivar index: bit number
        """
        libadc.adc_set_conf_mask(byref(self), index)

    def mask_set_from_device(self, adc):
        """
        Set the bit mask according to device capabilities

        :ivar adc: ADC device instance
        """
        libadc.adc_set_conf_mask_all(byref(self), adc)

    def value_set(self, index, value):
        """
        Set a value in the configuration structure

        :ivar index: configuration number
        :ivar value: configuration value
        """
        libadc.adc_set_conf(byref(self), index, value)

    def value_get(self, index):
        """
        Set a value in the configuration structure

        :ivar index: configuration number
        :return: the value (ctype.c_uint32)
        """
        value = c_uint32()
        libadc.adc_get_conf(byref(self), index, pointer(value))
        return value.value


class PyAdcBuf(Structure):
    """
    ADC data buffer

    :ivar data: sample buffer
    :ivar metadata: custom metadata
    :ivar samplesize: size of each sample in bits
    :ivar nsamples: number of samples in data
    :ivar dev: pointer to the ADC device that crated this buffer
    :ivar mapaddr:
    :ivar maplen:
    :ivar flags:
    """
    _fields_ = [("data", POINTER(c_short)),
                ("metadata", c_void_p),
                ("samplesize", c_int),
                ("nsamples", c_int),
                ("dev", c_void_p),
                ("mapaddr", c_void_p),
                ("maplen", c_uint),
                ("flags", c_uint)]
    @property
    def tstamp(self):
        """
        Extracts the timestamp

        :raises OSError: from C library errors
        :return: the timestamp (PyAdcLib.adc_timestamp)
        """
        ts = PyAdcTimestamp()
        libadc.adc_tstamp_buffer(pointer(self), pointer(ts))
        return ts

    def get_sample(self, chan, sample):
        """
        Get a sample from the buffer

        :ivar chan: channel number (0-based)
        :ivar sample: sample number (0-based)
        :return: the sample value (ctypes.c_int32)
        """
        value = c_int32()
        libadc.adc_buffer_get_sample(pointer(self), chan, sample, byref(value))
        return value.value


class timeval(Structure):
    """
    timeval structure

    :ivar sec: seconds
    :ivar usec: micro seconds
    """
    _fields_ = [("sec", c_long),
                ("usec", c_long)]


class PyAdcTimestamp(Structure):
    """
    ADC timestamp format

    :ivar sec: seconds
    :ivar ticks: 8ns increments
    :ivar bins: pico-seconds within each tick with 12bit resolution
                (bins * 8000/4096)
    """
    _fields_ = [("secs", c_ulong),
                ("ticks", c_ulong),
                ("bins", c_ulong)]

def libadc_create():
    """
    Initialize the libadc C library

    :return: a valid object to access the libadc library
    """
    def error_check_int(ret, func, args):
        """Generic error checker for functions returning 0 as success
        and -1 as error"""
        if ret < 0:
            raise OSError(get_errno(),
                          adc_strerror(get_errno()), "")
        else:
            return ret

    def error_check_pointer(ret, func, args):
        """Generic error handler for functions returning pointers"""
        if ret is None:
            raise OSError(get_errno(),
                          adc_strerror(get_errno()), "")
        else:
            return ret

    libadc = CDLL("libadc.so", use_errno=True)

    libadc.adc_strerror.argtypes = [c_int]
    libadc.adc_strerror.restype = c_char_p

    libadc.adc_print_version.restype = None
    libadc.adc_print_version.argtypes = None

    libadc.adc_init.restype = c_int
    libadc.adc_init.argtypes = None
    libadc.adc_init.errcheck = error_check_int

    libadc.adc_exit.restype = None
    libadc.adc_exit.argtypes = None

    libadc.adc_set_conf.restype = None
    libadc.adc_set_conf.argtypes = [POINTER(PyAdcConf), c_uint, c_uint]

    libadc.adc_set_conf_mask.restype = None
    libadc.adc_set_conf_mask.argtypes = [POINTER(PyAdcConf), c_uint]

    libadc.adc_get_conf.restype = c_int
    libadc.adc_get_conf.argtypes = [POINTER(PyAdcConf), c_uint,
                                    POINTER(c_uint32)]
    libadc.adc_get_conf.errcheck = error_check_int

    libadc.adc_get_driver_type.restype = c_char_p
    libadc.adc_get_driver_type.argtypes = [c_void_p]

    libadc.adc_open.restype = c_void_p
    libadc.adc_open.argtypes = [c_char_p, c_uint, c_ulong, c_uint, c_ulong]
    libadc.adc_open.errcheck = error_check_pointer

    libadc.adc_open_by_lun.restype = c_void_p
    libadc.adc_open_by_lun.argtypes = [c_char_p, c_int, c_ulong, c_uint,
                                     c_ulong]
    libadc.adc_open_by_lun.errcheck = error_check_pointer

    libadc.adc_request_buffer.restype = POINTER(PyAdcBuf)
    libadc.adc_request_buffer.argtypes = [c_void_p, c_int, c_void_p, c_uint]
    libadc.adc_request_buffer.errcheck = error_check_pointer

    libadc.adc_apply_config.restype = c_int
    libadc.adc_apply_config.argtypes = [c_void_p, c_uint, c_void_p]
    libadc.adc_apply_config.errcheck = error_check_int

    libadc.adc_acq_start.restype = c_int
    libadc.adc_acq_start.argtypes = [c_void_p, c_uint, c_void_p]
    libadc.adc_acq_start.errcheck = error_check_int

    libadc.adc_acq_poll.restype = c_int
    libadc.adc_acq_poll.argtypes = [c_void_p, c_uint, c_void_p]
    libadc.adc_acq_poll.errcheck = error_check_int

    libadc.adc_fill_buffer.restype = c_int
    libadc.adc_fill_buffer.argtypes = [c_void_p, c_void_p, c_uint, c_void_p]
    libadc.adc_fill_buffer.errcheck = error_check_int

    libadc.adc_acq_stop.restype = c_int
    libadc.adc_acq_stop.argtypes = [c_void_p, c_uint]
    libadc.adc_acq_stop.errcheck = error_check_int

    libadc.adc_tstamp_buffer.restype = POINTER(PyAdcTimestamp)
    libadc.adc_tstamp_buffer.argtypes = [c_void_p, c_void_p]
    libadc.adc_tstamp_buffer.errcheck = error_check_pointer

    libadc.adc_release_buffer.restype = c_int
    libadc.adc_release_buffer.argtypes = [c_void_p, c_void_p, c_void_p]
    libadc.adc_release_buffer.errcheck = error_check_int

    libadc.adc_close.restype = c_int
    libadc.adc_close.argtypes = [c_void_p]
    libadc.adc_close.errcheck = error_check_int

    libadc.adc_trigger_fire.restype = c_int
    libadc.adc_trigger_fire.argtypes = [c_void_p]
    libadc.adc_trigger_fire.errcheck = error_check_int

    libadc.adc_has_trigger_fire.restype = c_int
    libadc.adc_has_trigger_fire.argtypes = [c_void_p]

    """this function doesn't have body
libadc.adc_reset_conf.restype = c_int
libadc.adc_reset_conf.argtypes = [c_void_p, c_uint, c_void_p]
libadc.adc_reset_conf.errcheck = error_check_int"""

    libadc.adc_retrieve_config.restype = c_int
    libadc.adc_retrieve_config.argtypes = [c_void_p, c_void_p]
    libadc.adc_retrieve_config.errcheck = error_check_int

    libadc.adc_get_capabilities.restype = c_ulong
    libadc.adc_get_capabilities.argtypes = [c_void_p, c_int]
    libadc.adc_get_capabilities.errcheck = error_check_int

    libadc.adc_get_param.restype = c_int
    libadc.adc_get_param.argtypes = [c_void_p, c_char_p, c_char_p, c_void_p]
    libadc.adc_get_param.errcheck = error_check_int

    libadc.adc_set_param.restype = c_int
    libadc.adc_set_param.argtypes = [c_void_p, c_char_p, c_char_p, c_void_p]
    libadc.adc_set_param.errcheck = error_check_int

    libadc.adc_set_conf_mask_all.restype = None
    libadc.adc_set_conf_mask_all.argtypes = [c_void_p, c_void_p]

    libadc.adc_buffer_get_sample.restype = c_int
    libadc.adc_buffer_get_sample.argtypes = [c_void_p, c_uint, c_uint, POINTER(c_int32)]
    libadc.adc_buffer_get_sample.errcheck = error_check_int

    return libadc

libadc = libadc_create()

def adc_strerror(err):
    """
    Return ADC errors

    :ivar err: error number
    :return: an error string
    """
    return libadc.adc_strerror(err)

class PyAdcAbstract():
    """
    Abstract definition on and ADC
    """

    ADC_F_USERMASK = 0xffff0000
    ADC_F_FLUSH = 0x00010000
    ADC_F_VERBOSE = 0x00020000
    ADC_F_FIXUP = 0x00400000

    ADC_TRG_POL_POS = 0
    ADC_TRG_POL_NEG = 1

    def print_version(self):
        self.adc_print_version()

    def __init__(self, devid, board=None):
        """
        Initialize the ADC instance

        :ivar devid: ADC device ID
        """
        self.dev_id = devid
        self.__ADC_CONF_LEN = 64
        """number of allocated items in each structure"""

        if not hasattr(self, 'BOARD_NAME'):
            raise Exception("Missing board name")

        libadc.adc_init()
        self.tkn = self.__open(self.dev_id, 0, 0, self.ADC_F_FLUSH)

    def __del__(self):
        if hasattr(self, "tkn"):
            libadc.adc_close(self.tkn)
        libadc.adc_exit()

    def get_driver_type(self):
        return libadc.adc_get_driver_type(self.tkn)

    def __open(self, dev_id, totalsamples, nbuffer, flags):
        return libadc.adc_open(bytes(self.BOARD_NAME, "ascii"), dev_id,
                             totalsamples, nbuffer, flags)

    def __open_by_lun(self, name, lun, totalsamples, nbuffer, flags):
        return libadc.adc_open_by_lun(bytes(self.BOARD_NAME, "ascii"), lun,
                                    totalsamples, nbuffer, flags)

    def request_buffer(self, nsamples, alloc, flags):
        """
        Get a buffer instance

        :ivar nsamples: number of samples
        :ivar alloc: allocation function, use None unless you have very
                     special needs
        :ivar flags:
        :raises OSError: from C library errors
        :return: a buffer instance (PyAdcLib.PyAdcBuf)
        """
        return libadc.adc_request_buffer(self.tkn, nsamples, alloc, flags)

    def acq_start(self, flags, timeout_tv):
        """
        Start an acquisition

        :ivar flags:
        :ivar timeout_tv: timeout (PyAdcLib.timeval)
        :raises OSError: from C library errors
        """
        libadc.adc_acq_start(self.tkn, flags, byref(timeout_tv))

    def acq_poll(self, flags, timeout):
        """
        Poll the driver for buffers

        :ivar flags:
        :ivar timeout: timeout (PyAdcLib.timeval)
        :raises OSError: from C library errors
        """
        libadc.adc_acq_poll(self.tkn, flags, byref(timeout))

    def fill_buffer(self, buf, flags, timeout):
        """
        Fill a given buffer with samples

        :ivar buf: buffer (PyAdcLib.PyAdcBuf)
        :ivar flags:
        :ivar timeout: timeout (PyAdcLib.timeval)
        :raises OSError: from C library errors
        """
        libadc.adc_fill_buffer(self.tkn, buf, flags, byref(timeout))

    def acq_stop(self, flags):
        """
        Stop acquisition

        :ivar flags:
        :raises OSError: from C library errors
        """
        libadc.adc_acq_stop(self.tkn, flags)

    def release_buffer(self, buf, free):
        """
        Release a buffer instance

        :ivar buf: buffer (PyAdcLib.PyAdcBuf)
        :ivar free: memory release function, use None unless you used
                    an allocation function in request_buffer
        :ivar flags:
        :raises OSError: from C library errors
        :return: a buffer instance (PyAdcLib.PyAdcBuf)
        """
        libadc.adc_release_buffer(self.tkn, buf, free)

    def trigger_fire(self):
        """
        Force a trigger to fire (software trigger)

        :raises OSError: from C library errors
        """
        libadc.adc_trigger_fire(self.tkn)

    def has_trigger_fire(self):
        """
        Check if the software trigger is supported

        :return: true is supported, otherwise false
        """
        return libadc.adc_has_trigger_fire(self.tkn)

    def apply_config(self, config, flags):
        """
        Apply a configuration

        :ivar config: configuration (PyAdcLib.adc_conf)
        :ivar flags:
        :raises OSError: from C library errors
        """
        libadc.adc_apply_config(self.tkn, flags, byref(config))

    def retrieve_config(self, config):
        """
        Retrieve a configuration

        :ivar config: configuration (PyAdcLib.adc_conf)
        :raises OSError: from C library errors
        """
        libadc.adc_retrieve_config(self.tkn, byref(config))

    def get_capabilities(self, type):
        """
        Get a bitmask of the capabilities supported by a given type

        :ivar type: configuration type
        :return: a capability bitmask
        """
        libadc.adc_get_capabilities(self.tkn, type)

    def get_param(self, name, sptr, iptr):
        """
        Get a configuration value by name.
        This is mainly to handle custom values not supported directly by
        the adc_conf structure

        :ivar name: file name
        :ivar sptr: pointer to string to store the result as string
        :ivar iptr: pointer to integer to store the result as int
        :raises OSError: from C library errors
        """
        libadc.adc_get_param(self.tkn, name, sptr, iptr)

    def set_param(self, name, sptr, iptr):
        """
        Set a configuration value by name.
        This is mainly to handle custom values not supported directly by
        the adc_conf structure

        :ivar name: file name
        :ivar sptr: pointer to string to write a string
        :ivar iptr: pointer to integer to write an integer
        :raises OSError: from C library errors
        """
        libadc.adc_set_param(self.tkn, name, sptr, iptr)


class PyFmcAdc100m14b4ch(PyAdcAbstract):
    """
    Object for the FMC ADC 100M 14bit 4 channel device
    """
    BOARD_NAME = "fmc-adc-100m14b4cha"

    ADC_CONF_100M14B4CHA_CHN_RANGE_N = 3

    ADC_CONF_100M14B4CHA_CHN_RANGE_OPEN_DRAIN = 0
    ADC_CONF_100M14B4CHA_CHN_RANGE_100mV = 0x23
    ADC_CONF_100M14B4CHA_CHN_RANGE_1V = 0x11
    ADC_CONF_100M14B4CHA_CHN_RANGE_10V = 0x45
    ADC_CONF_100M14B4CHA_CHN_RANGE_100mV_CAL = 0x42
    ADC_CONF_100M14B4CHA_CHN_RANGE_1V_CAL = 0x40
    ADC_CONF_100M14B4CHA_CHN_RANGE_10V_CAL = 0x44

    ADC_CONF_100M14B4CHA_BUF_KMALLOC = 0
    ADC_CONF_100M14B4CHA_BUF_VMALLOC = 1

    ADC_CONF_100M14B4CHA_BUF_TYPE = 0
    ADC_CONF_100M14B4CHA_TRG_SW_EN = 1
    ADC_CONF_100M14B4CHA_ACQ_MSHOT_MAX = 2
    ADC_CONF_100M14B4CHA_BUF_SIZE_KB = 3
    ADC_CONF_100M14B4CHA_TRG_ALT_EN = 4
    __ADC_CONF_100M14B4CHA_LAST_INDEX = 5

    ADC_CONF_100M14B4CHA_FSM_STATE_IDLE = 1
    ADC_CONF_100M14B4CHA_FSM_STATE_PRE = 2
    ADC_CONF_100M14B4CHA_FSM_STATE_WAIT = 3
    ADC_CONF_100M14B4CHA_FSM_STATE_POST = 4
    ADC_CONF_100M14B4CHA_FSM_STATE_DECR = 5

    def __init__(self, devid):
        super(PyFmcAdc100m14b4ch, self).__init__(devid)
        self.__dbg_path = "/sys/kernel/debug/fmc-adc-100m.{:d}.auto/".format(self.dev_id)

    @property
    def pattern_data(self):
        with open(os.path.join(self.__dbg_path, "data_pattern"), 'r') as f:
            tmp = f.read().split()
        assert tmp[0] == "adc"
        return None if tmp[1] == "0" else int(tmp[2], 0)

    @pattern_data.setter
    def pattern_data(self, value):
        with open(os.path.join(self.__dbg_path, "data_pattern"), 'w') as f:
            if value is not None:
                f.write("adc 1 0x{:x}".format(value))
            else:
                f.write("adc 0")

class PyAdcGenericFake(PyAdcAbstract):
    """
    Generic device that does not need any hardware, nor driver.
    Used for testing
    """
    BOARD_NAME = "adc-genericfake"

    def __init__(self, devid):
        super(AdcGenericFake, self).__init__(devid)


class PyAdcZioFake(PyAdcAbstract):
    """
    Generic device that does not need any hardware.
    Used for testing the ZIO framework.
    """

    BOARD_NAME = "adc-ziofake"

    def __init__(self, devid):
        super(AdcZioFake, self).__init__(devid)
