/*
 * Routing public functions to device-specific code
 *
 * Copyright (C) 2018 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@cern.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */

#include <errno.h>
#include <string.h>

#include "adc-lib.h"
#include "adc-lib-int.h"

/**
 * It computes the avarege voltage within the given buffer
 * @param[in] buf data set to use
 * @param[in] chan channel number [0, NCHAN]
 * @param[out] avg the computer avarage. The scale depends on the
 *             configuration,
 * @return 0 on success, -1 on error and errno is set appropriately
 *         EINVAL: if the buffer is invalid, or channel is invalid
 */
int adc_buffer_math_avg(struct adc_buffer *buf,
			unsigned int chan,
			int32_t *avg)
{
	struct adc_conf cfg_brd;
	uint32_t nchan = 0;
	int i, err;
	int64_t total = 0;


	memset(&cfg_brd, 0, sizeof(struct adc_conf));
	cfg_brd.type = ADC_CONF_TYPE_BRD;
	adc_set_conf_mask_all(&cfg_brd, buf->dev);
	err = adc_retrieve_config(buf->dev, &cfg_brd);
	if (err)
		return err;
	adc_get_conf(&cfg_brd, ADC_CONF_BRD_N_CHAN, &nchan);

	if (chan >= nchan) {
		errno = EINVAL;
		return -1;
	}

	for (i = 0; i < buf->nsamples; ++i) {
		int32_t val;

		err = adc_buffer_get_sample(buf, chan, i, &val);
		if (err)
			return err;
		total += val;
	}
	*avg = total / buf->nsamples;

	return 0;
}
