/*
 * Routing public functions to device-specific code
 *
 * Copyright (C) 2013 CERN (www.cern.ch)
 * Author: Federico Vaga <federico.vaga@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2 as published by the Free Software Foundation or, at your
 * option, any later version.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "adc-lib.h"
#include "adc-lib-int.h"

#define PATH_MAX 1024

/**
 * List of supported board names
 */
const char *adc_board_name[] = {
	[FMCADC_100MS_4CH_14BIT] = ADC_NAME_FMCADC100M,
	[ADC_ZIOFAKE] = ADC_NAME_ZIOFAKE,
	[ADC_GENERICFAKE] = ADC_NAME_GENERICFAKE,
};

/*
 * The following array is the main entry point into the boards
 */
static const struct adc_board_type *adc_board_types[] = {
	[FMCADC_100MS_4CH_14BIT] = &fmcadc_100ms_4ch_14bit,
	[ADC_ZIOFAKE] = &adc_ziofake,
	[ADC_GENERICFAKE] = &adc_genericfake,
	/* add new boards here */
};


static const struct adc_board_type *find_board(char *name)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(adc_board_types); i++)
		if (adc_board_types[i] && !strcmp(name, adc_board_types[i]->name))
			return adc_board_types[i];
	errno = ENODEV;
	return NULL;
}


/**
 * It opens a link with an ADC device
 * @param[in] name name of the board type to open according to adc_board_name.
 * @param[in] dev_id The device identifier is used to enumerate several
 *     cards in the same system.  The number is usually dependent on the
 *     geographic placement of the card (bus number, slot number).
 * @param[in] totalsamples This is a a hint about how big a buffer the
 *     application will use. i.e. ``nshots * (presamples + postsamples)``
 * @param[in] nbuffer This is a hint about how many buffers are being used
 *     at the same time by the application. For example, a multi-shot
 *     acquisition requires all buffers to be available at the same time.
 *     Again, this information may require device-level configuration at
 *     open time.
 * @param[in] flags currently it supports: ADC_F_FLUSH, ADC_F_VERBOSE
 *
 * @return It returns an opaque object describing the device.
 *         It should be passed around but not be looked into.
 *
 * @todo Open should choose the buffer type
 *
 * It initialize the board, this can be different board to board.
 */
struct adc_dev *adc_open(char *name, unsigned int dev_id,
			 unsigned long totalsamples,
			 unsigned int nbuffer,
			 unsigned long flags)
{
	const struct adc_board_type *b;

	b = find_board(name);
	if (!b)
		return NULL;

	if (!b->adc_op->open) {
		errno = ADC_ENOP;
		return NULL;
	}
	return b->adc_op->open(b, dev_id, totalsamples, nbuffer, flags);
}



#define ADC_PATH_PATTERN "/dev/%s.%d"

/**
 * It opens a link with an ADC device using a unit number identifier
 * @param[in] name name of the board type to open according to adc_board_name.
 * @param[in] lun the Logical Unit Number. This number comes from a database
 *     description of the system.
 * @param[in] buffersize This is a a hint about how big a buffer the
 *     application will use. i.e. ``nshots * (presamples + postsamples)``
 * @param[in] nbuffer This is a hint about how many buffers are being used
 *     at the same time by the application.
 * @param[in] flags currently it supports: ADC_F_FLUSH, ADC_F_VERBOSE
 *
 * @return It returns an opaque object describing the device.
 *         It should be passed around but not be looked into.
 *
 * @todo make it generic, now it works only for the CERN environment
 *       with the adc-100m14b. Actually I (Federico) believe that this
 *       is impossible to make it generic enough
 * @todo some udev rule should create symlink to map lun->devid
 */
struct adc_dev *adc_open_by_lun(char *name, int lun,
				unsigned long buffersize,
				unsigned int nbuffer,
				unsigned long flags)
{
	ssize_t ret;
	char dev_id_str[8];
	char path[PATH_MAX];
	uint32_t dev_id;

	ret = snprintf(path, sizeof(path), "/dev/%s.%d",
		       "adc-100m14b" /* FIXME: this must be generic */,
		       lun);
	if (ret < 0 || ret >= sizeof(path)) {
		errno = EINVAL;
		return NULL;
	}
	ret = readlink(path, dev_id_str, sizeof(dev_id_str));
	if (ret < 0)
		return NULL;
	if (sscanf(dev_id_str, "%4x", &dev_id) != 1) {
		errno = ENODEV;
		return NULL;
	}
	return adc_open(name, dev_id, buffersize, nbuffer, flags);

}


/**
 * It closes the link with the ADC devices previously opened with adc_open()
 * @param[in] dev the device token from adc_open()
 * @return It returns 0 on success, otherwise -1 and errno is appropriately set
 */
int adc_close(struct adc_dev *dev)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->close) {
		errno = ADC_ENOP;
		return -1;
	}
	return b->adc_op->close(dev);
}


/**
 * It tells hardware to start the acquisition on the given device according
 * to the current configuration (but starting in itself is usually a fast
 * operation, because it doesn't rewrite configuration to the hardware,
 * at least with currently supported boards).
 * It can return immediately or wait for completion, with a timeout.
 * @param[in] dev ADC device token
 * @param[in] flags flags to control the acquisition options
 *     ADC_F_FLUSH - discard prevoius leftover data before activating
 *                   the acquisition (if any is there)
 * @param[in] timeout The timeout is used like the argument for ``select(2)``:
 *     if NULL, the functions wait forever, otherwise they specify the
 *     maximum allowed waiting time -- and passing zero values asks to not wait
 *     at all (you can use adc_acq_poll() later).
 *
 * @return 0 on success, -1 on error and errno is set appropriately
 *     ADC_EDISABLED - acquisition disabled or aborted
 */
int adc_acq_start(struct adc_dev *dev,
		  unsigned int flags,
		  struct timeval *timeout)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->acq_start) {
		errno = ADC_ENOP;
		return -1;
	}

	return b->adc_op->acq_start(dev, flags, timeout);
}


/**
 * It waits for acquisition to complete.
 * It polls the given device to check if data is available
 * @param[in] dev ADC device token
 * @param[in] flags flags to specify how to poll
 * @param[in] timeout The timeout is used like the argument for ``select(2)``:
 *     if NULL, the functions wait forever, otherwise they specify the
 *     maximum allowed waiting time -- and passing zero values asks to not wait
 *     at all.
 * @return 0 on success, -1 on error and errno is set appropriately
 *     ADC_EDISABLED - acquisition disabled or aborted
 *     EAGAIN - time out
 *
 * @todo For multi-shot we would like to return the number of
 *     still-missing shots, but this is not yet supported.
 */
int adc_acq_poll(struct adc_dev *dev, unsigned int flags,
		 struct timeval *timeout)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->acq_poll) {
		errno = ADC_ENOP;
		return -1;
	}

	return b->adc_op->acq_poll(dev, flags, timeout);
}


/**
 * It stops an already-begun acquisition. It is not expected to be of
 * frequent use.
 * @param[in] dev ADC device token
 * @param[in] flags flags to specify how to stop
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int adc_acq_stop(struct adc_dev *dev, unsigned int flags)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->acq_stop) {
		errno = ADC_ENOP;
		return -1;
	}

	return b->adc_op->acq_stop(dev, flags);
}

/**
 * Flush any acquired data from any software or hardware buffer
 * @param[in] dev ADC device token
 * @param[in] flags flags for fine control
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int adc_acq_flush(struct adc_dev *dev, unsigned int flags)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->acq_flush) {
		errno = ADC_ENOP;
		return -1;
	}

	if (flags & ~ADC_FLUSH_F_ALL) {
		errno = ADC_ENOFLAGS;
		return -1;
	}


	return b->adc_op->acq_flush(dev, flags);
}

/**
 * It applies the given configuration to the device
 * @param[in] dev ADC device token
 * @param[in] flags
 * @param[in] conf configuration to apply
 * @return 0 on success, -1 on error and errno is set appropriately
 *
 * If the configuration cannot be applied entirely, then this function
 * will set ADC_CONF_F_ERROR in conf->flags
 */
int adc_apply_config(struct adc_dev *dev, unsigned int flags,
		     struct adc_conf *conf)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;
	uint64_t cap_mask;
	int err;

	conf->flags &= ~ADC_CONF_F_ERROR;

	if (!b->adc_op->apply_config) {
		errno = ADC_ENOP;
		return -1;
	}

	if (conf->type >= __ADC_CONF_TYPE_LAST_INDEX) {
		errno = ADC_ENOCFG;
		return -1;
	}

	if (!conf->mask) {
		errno = ADC_ENOMASK; /* Just for notification */
		return  0; /* Nothing to do */
	}
	cap_mask = b->capabilities[conf->type] & conf->mask;
	if (cap_mask != conf->mask) {
		/* Unsupported capabilities */
		errno = ADC_ENOCAP;
		return -1;
	}
	err = b->adc_op->apply_config(dev, flags, conf);
	if (err)
		conf->flags |= ADC_CONF_F_ERROR;
	return err;

}

/**
 * It applies the given configuration to the device
 * @param[in] dev ADC device token
 * @param[in] flags
 * @param[in] conf list of configurations to apply
 * @param[in] n number of valid configurations in conf
 * @return 0 on success, -1 on error and errno is set appropriately
 *
 * If a configuration cannot be applied entirely, then this function
 * will set ADC_CONF_F_ERROR in conf[i].flags
 */
int adc_apply_config_n(struct adc_dev *dev, unsigned int flags,
		       struct adc_conf *conf, unsigned int n)
{
	unsigned int i;
	int err_last = 0;

	for (i = 0; i < n; ++i) {
		int err = adc_apply_config(dev, flags, &conf[i]);
		if (err) {
			err_last = err;
			continue;
		}
	}

	return err_last;
}

/**
 * It retrieve the current device configuration
 * @param[in] dev ADC device token
 * @param[in,out] conf configuration descriptor to fill with configuration
 *                values.
 * @return 0 on success, -1 on error and errno is set appropriately
 *
 * If the configuration cannot be retrieved entirely, then this function
 * will set ADC_CONF_F_ERROR in conf->flags
 */
int adc_retrieve_config(struct adc_dev *dev, struct adc_conf *conf)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;
	uint64_t cap_mask;
	int err;

	conf->flags &= ~ADC_CONF_F_ERROR;

	if (!b->adc_op->retrieve_config) {
		errno = ADC_ENOP;
		return -1;
	}

	if (conf->type >= __ADC_CONF_TYPE_LAST_INDEX) {
		errno = ADC_ENOCFG;
		return -1;
	}

	if (!conf->mask) {
		errno = ADC_ENOMASK; /* Just for notification */
		return 0; /* Nothing to do */
	}
	cap_mask = b->capabilities[conf->type] & conf->mask;
	if (cap_mask != conf->mask) {
		/* Unsupported capabilities */
		errno = ADC_ENOCAP;
		return -1;
	}
	err = b->adc_op->retrieve_config(dev, conf);
	if (err)
		conf->flags |= ADC_CONF_F_ERROR;
	return err;
}

/**
 * It retrieve the current device configuration
 * @param[in] dev ADC device token
 * @param[in,out] conf configuration descriptor to fill with configuration
 *                values.
 * @param[in] n number of valid configurations in conf
 * @return 0 on success, -1 on error and errno is set appropriately
 *
 * If a configuration cannot be retrieved entirely, then this function
 * will set ADC_CONF_F_ERROR in conf[i].flags
 */
int adc_retrieve_config_n(struct adc_dev *dev, struct adc_conf *conf,
			  unsigned int n)
{
	unsigned int i;
	int err_last = 0;

	for (i = 0; i < n; ++i) {
		int err = adc_retrieve_config(dev, &conf[i]);
		if (err) {
			err_last = err;
			continue;
		}
	}

	return err_last;
}

/**
 * It get a single parameter from the device
 * @param[in] dev ADC device token
 * @param[in] name parameter name to set
 * @param[out] sptr a string value (higher priority than iptr)
 * @param[out] iptr an integer value
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int adc_get_param(struct adc_dev *dev, char *name,
		     char *sptr, int *iptr)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->get_param) {
		errno = ADC_ENOP;
		return -1;
	}

	return b->adc_op->get_param(dev, name, sptr, iptr);
}


/**
 * It set a single parameter to the device
 * @param[in] dev ADC device token
 * @param[in] name parameter name to set
 * @param[in] sptr a string value (higher priority than iptr)
 * @param[in] iptr an integer value
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int adc_set_param(struct adc_dev *dev, char *name,
		     char *sptr, int *iptr)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->set_param) {
		errno = ADC_ENOP;
		return -1;
	}

	return b->adc_op->set_param(dev, name, sptr, iptr);
}


/**
 * It allocates an empty buffer where store data from the acquisition
 * @param[in] dev ADC device token
 * @param[in] nsamples size of the buffer in number of samples
 * @param[in] alloc when set it will use the given function to allocate
 *            the buffer
 * @param[in] flags flags to control the buffer request
 * @return It returns a valid pointer on success, otherwise NULL and errno
 *     is set appropriately
 */
struct adc_buffer *adc_request_buffer(struct adc_dev *dev,
					    int nsamples,
					    void *(*alloc)(size_t),
					    unsigned int flags)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->request_buffer) {
		errno = ADC_ENOP;
		return NULL;
	}

	return b->adc_op->request_buffer(dev, nsamples, alloc, flags);
}


/**
 * It ask the library to get the data from the device and fill the
 * given buffer.
 * @param[in] dev ADC device token
 * @param[in] buf the buffer to fill
 * @param[in] flags used to control how to fill the buffer
 *            (ADC_F_FIXUP)
 * @param[in] timeout maximum time to fill the buffer. The behavior
 *     is similar to *select(1)*. If NULL, there is no timeout and
 *     the function will wait until the data is ready.
 * @return 0 on success, -1 on error and errno is set appropriately
 *   ADC_EDISABLED - when the acquisition is disabled or aborted
 *   ETIMEDOUT - timeout
 *
 * The size of the buffer must be at least equal to the size of the shot
 * that library should copy into; otherwise, the result is undefined.
 */
int adc_fill_buffer(struct adc_dev *dev,
		       struct adc_buffer *buf,
		       unsigned int flags,
		       struct timeval *timeout)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;
	int err;

	if (!b->adc_op->fill_buffer) {
		errno = ADC_ENOP;
		return -1;
	}

	err = b->adc_op->fill_buffer(dev, buf, flags, timeout);
	if (err)
		return err;

	if (flags & ADC_F_FIXUP)
		return adc_buffer_fixup(buf);

	return 0;
}


/**
 * It gets the timestamp of a given buffer, in a driver-specific way
 * (most likely looking in the metadata structure).
 * @param[in] buf the buffer of interest
 * @param[out] ts the buffer timestamp
 * @return the buffer timestamp
 *
 * If the caller passes a ``ts`` pointer, the timestamp is copied into
 * that pointer and the same pointer is returned. If ``ts`` is NULL,
 * the function returns a pointer of its own choice, that is only valid
 * until the buffer is released.
 */
struct adc_timestamp *adc_tstamp_buffer(struct adc_buffer *buf,
					struct adc_timestamp *ts)
{
	struct adc_gid *g = (struct adc_gid *)buf->dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->tstamp_buffer) {
		errno = ADC_ENOP;
		return NULL;
	}

	return b->adc_op->tstamp_buffer(buf, ts);
}


/**
 * It releases any resources associated with the buffer.
 * @param[in] dev ADC device token
 * @param[in] buf the buffer to fill
 * @param free when set it will use the given function to release
 *     the buffer
 * @return 0 on success, -1 on error and errno is set appropriately
 */
int adc_release_buffer(struct adc_dev *dev, struct adc_buffer *buf,
			  void (*free)(void *))
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->release_buffer) {
		errno = ADC_ENOP;
		return -1;
	}

	if (!buf)
		return 0;

	return b->adc_op->release_buffer(dev, buf, free);
}

/**
 * It checks if the board support software trigger
 * @param[in] dev ADC device token
 * @return 1 when it does support software trigger; 0 when it does not
 */
int adc_has_trigger_fire(struct adc_dev *dev)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	return !!b->adc_op->trigger_fire;
}


/**
 * It forces the board to trigger the acquisition
 * @param[in] dev ADC device token
 * @return 0 on success, -1 on error and errno is set appropriately
 *         ADC_ENOP_SWTRG: when software trigger is not supported
 */
int adc_trigger_fire(struct adc_dev *dev)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->trigger_fire) {
		errno = ADC_ENOP_SWTRG;
		return -1;
	}
	return b->adc_op->trigger_fire(dev);
}

/**
 * It gets a sample from the buffer
 * @param[in] buf buffer to use
 * @param[in] chan which channel
 * @param[in] acq_sample acquisition sample
 * @param[out] value
 * @return 0 on success, -1 on error and errno is set appropriately
 *         ADC_ENOP if the operation is not supported by the board that
 *                  acquired the buffer
 */
int adc_buffer_get_sample(struct adc_buffer *buf,
			  unsigned int chan,
			  unsigned int acq_sample,
			  int32_t *value)
{
	struct adc_gid *g = (struct adc_gid *)buf->dev;
	const struct adc_board_type *b = g->board;

	if (!b->adc_op->buffer_get_sample) {
		errno = ADC_ENOP;
		return -1;
	}

	if (acq_sample >= buf->nsamples) {
		errno = EINVAL;
		return -1;
	}

	return b->adc_op->buffer_get_sample(buf, chan, acq_sample, value);
}

/**
 * It fixes the given buffer if there is any need
 * @param[in, out] buf acquistion buffer (filled)
 * @return 0 on success, -1 on error and errno is set appropriately
 *
 * This is used by board specifc code to compensate some known error
 */
int adc_buffer_fixup(struct adc_buffer *buf)
{
	struct adc_gid *g = (struct adc_gid *)buf->dev;
	const struct adc_board_type *b = g->board;

	if (b->adc_op->buffer_fixup)
		return b->adc_op->buffer_fixup(buf);
	return 0;
}

/**
 * It checks if the board support offset auto-clear
 * @param[in] dev ADC device token
 * @return 1 when it does support offset auto clear; 0 when it does not
 */
int adc_has_offset_auto_clear(struct adc_dev *dev)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	return !!b->adc_op->offset_auto_clear;
}

/**
 * It clears eventual offsets on all channel
 * @param[in] dev ADC device token
 * @param[in] flags options @see adc_offset_auto_clear_flags
 * @return 0 on success, -1 on error and errno is set appropriately
 *         EINVAL: invalid flags value
 *         ADC_ENOP_OFFCLR: when offset auto-clear is not supported
 *
 * NOTE: The function may overwrite your current configuration (unless
 * you use a flag) and it may remove any trace of previous acquisitions.
 *
 * Offset configuration is always overwritten (no matter what flag you use)
 */
int adc_offset_auto_clear(struct adc_dev *dev,
			  unsigned long flags)
{
	struct adc_gid *g = (struct adc_gid *)dev;
	const struct adc_board_type *b = g->board;

	if (!adc_has_offset_auto_clear(dev)) {
		errno = ADC_ENOP_OFFCLR;
		return -1;
	}

	if (flags & ~__ADC_OFFSET_AC_F_MASK) {
		errno = EINVAL;
		return -1;
	}

	return b->adc_op->offset_auto_clear(dev, flags);
}
