"""
SPDX-License-Identifier: GPL-3.0-or-later
SPDX-FileCopyrightText: 2020 CERN
"""

import pytest
import time
from PyAdcLib import PyAdcConf, PyFmcAdc100m14b4ch, timeval

@pytest.fixture(scope="function", params=[1 << i for i in range(14)] + \
                                         [0x2000 + (1 << i) for i in range(13)])
def adc_simple_pattern(adc_simple, request):
    adc_simple.pattern_data = request.param
    yield adc_simple
    adc_simple.pattern_data = None

class TestAdcAcquisitionPattern(object):

    def test_adc_acq_poll_fail(self, adc_simple):
        """Test that the poll fails with timeout if there are no triggers"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_ACQ, 0)
        conf.value_set(PyAdcConf.ADC_CONF_ACQ_N_SHOTS, 1)
        adc_simple.apply_config(conf, 0)

        timeout = timeval(0,0)
        adc_simple.acq_start(PyFmcAdc100m14b4ch.ADC_F_FLUSH,
                                         timeout)
        with pytest.raises(OSError):
            timeout = timeval(2,0)
            adc_simple.acq_poll(0, timeout)

    @pytest.mark.parametrize("nshots", range(1, 10))
    def test_adc_acq_poll(self, adc_simple, nshots):
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_ACQ, 0)
        conf.value_set(PyAdcConf.ADC_CONF_ACQ_N_SHOTS, nshots)
        adc_simple.apply_config(conf, 0)

        timeout = timeval(0,0)
        adc_simple.acq_start(PyFmcAdc100m14b4ch.ADC_F_FLUSH,
                                         timeout)
        for i in range(nshots):
            adc_simple.trigger_fire()
            time.sleep(0.1)
        timeout = timeval(10,0)
        adc_simple.acq_poll(0, timeout)

    @pytest.mark.parametrize("pre_samples,post_samples",
                             [(10**x, 10**y) for x in range(1, 3) for y in range(1, 4)])
    @pytest.mark.parametrize("nshots", range(1, 10))
    def test_adc_acq_basic(self, adc_simple_pattern,
                           pre_samples, post_samples, nshots):
        """Test that we can get"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_ACQ, 0)
        conf.value_set(PyAdcConf.ADC_CONF_ACQ_PRE_SAMP, pre_samples)
        conf.value_set(PyAdcConf.ADC_CONF_ACQ_POST_SAMP, post_samples)
        conf.value_set(PyAdcConf.ADC_CONF_ACQ_N_SHOTS, nshots)
        adc_simple_pattern.apply_config(conf, 0)

        nsamples = pre_samples + post_samples
        buf = adc_simple_pattern.request_buffer(nsamples, None, 0)
        adc_simple_pattern.acq_start(PyFmcAdc100m14b4ch.ADC_F_FLUSH,
                                         timeval(0, 0))
        for n in range(nshots):
            time.sleep(0.5 + 1e-8 * (pre_samples + post_samples))
            adc_simple_pattern.trigger_fire()

        pattern = adc_simple_pattern.pattern_data
        for n in range(nshots):
            adc_simple_pattern.acq_poll(0, timeval(10, 0))
            adc_simple_pattern.fill_buffer(buf, 0, timeval(10, 0))
            assert buf.contents.nsamples == nsamples

            for chan in range(4):
                for s in range(nsamples):
                    sample = abs(buf.contents.get_sample(chan, s) >> 2)
                    if pattern >= 0x2000:
                        # for negative numbers
                        assert 0x4000 == pattern + sample, str(buf.contents.data[:nsamples * 4])
                    else:
                        assert sample == pattern, str(buf.contents.data[:nsamples * 4])

        adc_simple_pattern.release_buffer(buf, None)
