"""
SPDX-License-Identifier: GPL-3.0-or-later
SPDX-FileCopyrightText: 2020 CERN
"""

import pytest
import ctypes
import random
from PyAdcLib import PyAdcConf, PyFmcAdc100m14b4ch


class TestAdcGetterSetterBoard(object):

    def test_adc_state_machine_status(self, adc100m14b4cha):
        """Test that we get an IDLE state machine since we are not running
        any acquisition"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_BRD, 0)
        conf.mask_set(PyAdcConf.ADC_CONF_BRD_STATE_MACHINE_STATUS)
        adc100m14b4cha.retrieve_config(conf)
        assert conf.value_get(PyAdcConf.ADC_CONF_BRD_STATE_MACHINE_STATUS) == 1

    def test_adc_n_chan(self, adc100m14b4cha):
        """Test that we can get the number of channels"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_BRD, 0)
        conf.mask_set(PyAdcConf.ADC_CONF_BRD_N_CHAN)
        adc100m14b4cha.retrieve_config(conf)
        assert conf.value_get(PyAdcConf.ADC_CONF_BRD_N_CHAN) == 4

    def test_adc_n_trg_ext(self, adc100m14b4cha):
        """Test that we get the number of external triggers"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_BRD, 0)
        conf.mask_set(PyAdcConf.ADC_CONF_BRD_N_TRG_EXT)
        adc100m14b4cha.retrieve_config(conf)
        assert conf.value_get(PyAdcConf.ADC_CONF_BRD_N_TRG_EXT) == 1

    def test_adc_n_trg_thr(self, adc100m14b4cha):
        """Test that we get the number of threshold triggers"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_BRD, 0)
        conf.mask_set(PyAdcConf.ADC_CONF_BRD_N_CHAN)
        conf.mask_set(PyAdcConf.ADC_CONF_BRD_N_TRG_THR)
        adc100m14b4cha.retrieve_config(conf)
        assert conf.value_get(PyAdcConf.ADC_CONF_BRD_N_TRG_THR) == 4
        assert conf.value_get(PyAdcConf.ADC_CONF_BRD_N_CHAN) == 4

    def test_adc_n_trg_ext(self, adc100m14b4cha):
        """Test that we get the number of timer triggers"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_BRD, 0)
        conf.mask_set(PyAdcConf.ADC_CONF_BRD_N_TRG_TIM)
        adc100m14b4cha.retrieve_config(conf)
        assert conf.value_get(PyAdcConf.ADC_CONF_BRD_N_TRG_TIM) == 1


class TestAdcGetterSetterChannel(object):

    @pytest.mark.parametrize("vref", [0x45, 0x11, 0x23])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_vref(self, adc100m14b4cha, vref, channel):
        """Test that we can read/write vref on all channels"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, channel)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_RANGE, vref)
        assert conf.value_get(PyAdcConf.ADC_CONF_CHN_RANGE) == ctypes.c_uint32(vref).value
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_CHN_RANGE)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_CHN_RANGE) == conf_rb.value_get(PyAdcConf.ADC_CONF_CHN_RANGE)

    @pytest.mark.parametrize("termination", [True, False])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_termination(self, adc100m14b4cha, termination, channel):
        """Test that we can read/write termination on all channels"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, channel)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_TERMINATION, int(termination))
        assert conf.value_get(PyAdcConf.ADC_CONF_CHN_TERMINATION) == int(termination)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_CHN_TERMINATION)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_CHN_TERMINATION) == conf_rb.value_get(PyAdcConf.ADC_CONF_CHN_TERMINATION)

    @pytest.mark.parametrize("offset", range(-5000000, 5000000 + 1, 100000))
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_offset(self, adc100m14b4cha, offset, channel):
        """Test that we can read/write offset on all channels"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, channel)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_OFFSET, offset)
        assert conf.value_get(PyAdcConf.ADC_CONF_CHN_OFFSET) == ctypes.c_uint32(offset).value
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_CHN_OFFSET)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_CHN_OFFSET) == conf_rb.value_get(PyAdcConf.ADC_CONF_CHN_OFFSET)

    @pytest.mark.parametrize("offset", [-5000001, 5000001])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_offset_invalid(self, adc100m14b4cha, offset, channel):
        """Test that we can read/write offset on all channels"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, channel)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_OFFSET, offset)
        assert conf.value_get(PyAdcConf.ADC_CONF_CHN_OFFSET) == ctypes.c_uint32(offset).value
        with pytest.raises(OSError):
            adc100m14b4cha.apply_config(conf, 0)

    @pytest.mark.parametrize("saturation", range(0, 0x7FFF, 1000))
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_saturation(self, adc100m14b4cha, saturation, channel):
        """Test that we can read/write saturation on all channels"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, channel)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_SATURATION, saturation)
        assert conf.value_get(PyAdcConf.ADC_CONF_CHN_SATURATION) == ctypes.c_uint32(saturation).value
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_CHN_SATURATION)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_CHN_SATURATION) == conf_rb.value_get(PyAdcConf.ADC_CONF_CHN_SATURATION)

    @pytest.mark.parametrize("saturation", [random.randrange(0x8000, 0xFFFFFFFF) for i in range(10)])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_saturation_invalid(self, adc100m14b4cha, saturation, channel):
        """Test that we can read/write saturation on all channels"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_CHN, channel)
        conf.value_set(PyAdcConf.ADC_CONF_CHN_SATURATION, saturation)
        assert conf.value_get(PyAdcConf.ADC_CONF_CHN_SATURATION) == ctypes.c_uint32(saturation).value
        with pytest.raises(OSError):
            adc100m14b4cha.apply_config(conf, 0)


class TestAdcGetterSetterTriggerThreshold(object):

    @pytest.mark.parametrize("enable", [True, False])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_enable(self, adc100m14b4cha, enable, channel):
        """Test that we can enable and disable a trigger threshold"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_ENABLE, int(enable))
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_ENABLE) == int(enable)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_TRG_THR_ENABLE)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_ENABLE) == conf_rb.value_get(PyAdcConf.ADC_CONF_TRG_THR_ENABLE)

    @pytest.mark.parametrize("polarity", [0, 1])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_polarity(self, adc100m14b4cha, polarity, channel):
        """Test that we can set the polarity"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_POLARITY, int(polarity))
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_POLARITY) == int(polarity)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_TRG_THR_POLARITY)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_POLARITY) == conf_rb.value_get(PyAdcConf.ADC_CONF_TRG_THR_POLARITY)

    @pytest.mark.parametrize("delay", [2**x for x in range(32)])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_delay(self, adc100m14b4cha, delay, channel):
        """Test that we set the trigger delay"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_DELAY, int(delay))
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_DELAY) == int(delay)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_TRG_THR_DELAY)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_DELAY) == conf_rb.value_get(PyAdcConf.ADC_CONF_TRG_THR_DELAY)

    @pytest.mark.parametrize("threshold", [2**x for x in range(15)])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_threshold(self, adc100m14b4cha, threshold, channel):
        """Test that we set the trigger threshold"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_THRESHOLD, int(threshold))
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_THRESHOLD) == int(threshold)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_TRG_THR_THRESHOLD)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_THRESHOLD) == conf_rb.value_get(PyAdcConf.ADC_CONF_TRG_THR_THRESHOLD)

    @pytest.mark.parametrize("threshold", [2**x for x in range(16, 32)])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_threshold_invalid(self, adc100m14b4cha, threshold, channel):
        """Test that we can't set an invalid the trigger threshold"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_THRESHOLD, int(threshold))
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_THRESHOLD) == int(threshold)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_TRG_THR_THRESHOLD)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert 0 == conf_rb.value_get(PyAdcConf.ADC_CONF_TRG_THR_THRESHOLD)

    @pytest.mark.parametrize("hysteresis", [2**x for x in range(15)])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_hysteresis(self, adc100m14b4cha, hysteresis, channel):
        """Test that we set the trigger hysteresis"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_HYSTERESIS, int(hysteresis))
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_HYSTERESIS) == int(hysteresis)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_TRG_THR_HYSTERESIS)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_HYSTERESIS) == conf_rb.value_get(PyAdcConf.ADC_CONF_TRG_THR_HYSTERESIS)

    @pytest.mark.parametrize("hysteresis", [2**x for x in range(16, 32)])
    @pytest.mark.parametrize("channel", range(4))
    def test_adc_hysteresis_invalid(self, adc100m14b4cha, hysteresis, channel):
        """Test that values greater that 16 bits are masked"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_THR_HYSTERESIS, int(hysteresis))
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_THR_HYSTERESIS) == int(hysteresis)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_THR, channel)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_TRG_THR_HYSTERESIS)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert 0 == conf_rb.value_get(PyAdcConf.ADC_CONF_TRG_THR_HYSTERESIS)

class TestAdcGetterSetterTriggerExternal(object):

    @pytest.mark.parametrize("enable", [True, False])
    def test_adc_enable(self, adc100m14b4cha, enable):
        """Test that we can enable and disable a trigger external"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_EXT, 0)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_EXT_ENABLE, int(enable))
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_EXT_ENABLE) == int(enable)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_EXT, 0)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_TRG_EXT_ENABLE)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_EXT_ENABLE) == conf_rb.value_get(PyAdcConf.ADC_CONF_TRG_EXT_ENABLE)

    @pytest.mark.parametrize("polarity", [0, 1])
    def test_adc_polarity(self, adc100m14b4cha, polarity):
        """Test that we can set the polarity"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_EXT, 0)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_EXT_POLARITY, int(polarity))
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_EXT_POLARITY) == int(polarity)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_EXT, 0)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_TRG_EXT_POLARITY)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_EXT_POLARITY) == conf_rb.value_get(PyAdcConf.ADC_CONF_TRG_EXT_POLARITY)

    @pytest.mark.parametrize("delay", [2**x for x in range(32)])
    def test_adc_delay(self, adc100m14b4cha, delay):
        """Test that we set the trigger delay"""
        conf = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_EXT, 0)
        conf.value_set(PyAdcConf.ADC_CONF_TRG_EXT_DELAY, int(delay))
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_EXT_DELAY) == int(delay)
        adc100m14b4cha.apply_config(conf, 0)

        conf_rb = PyAdcConf(PyAdcConf.ADC_CONF_TYPE_TRG_EXT, 0)
        conf_rb.mask_set(PyAdcConf.ADC_CONF_TRG_EXT_DELAY)
        adc100m14b4cha.retrieve_config(conf_rb)
        assert conf.value_get(PyAdcConf.ADC_CONF_TRG_EXT_DELAY) == conf_rb.value_get(PyAdcConf.ADC_CONF_TRG_EXT_DELAY)
